"""
Module for working with GluonTS and models from it
"""
import itertools
from pathlib import Path
from gluonts.trainer import Trainer
from gluonts.model.deepar import DeepAREstimator
from sklearn import preprocessing
import pandas as pd
import streamlit as st
import numpy as np

# def clean_duplicates(data, streamlit=False):
#     unique = data.loc[:, 'name'].unique()
#     num_unique = len(unique)
#     clean = pd.DataFrame()
#     for player in unique:
#         player_df = data.loc[data.loc[:, 'name'] == player]
#         player_df = player_df.drop_duplicates(subset='date')
#         clean = pd.concat([clean, player_df])
#     return clean, num_unique

def clean_duplicates(data, streamlit=False):
    unique = data.loc[:, 'name'].unique()
    clean = pd.DataFrame()
    for player in unique:
        player_df = data.loc[data.loc[:, 'name'] == player]
        player_df = player_df.drop_duplicates(subset='date')
        clean = pd.concat([clean, player_df])
    return clean

def clean_rookies_retirees(data, split_from='2018-10-03'):
    unique = data.loc[:, 'name'].unique()
    clean = pd.DataFrame()
    for player in unique:
        player_df = data.loc[data.loc[:, 'name'] == player]
        train_df = player_df.loc[player_df.loc[:, 'date'] < split_from]
        test_df = player_df.loc[player_df.loc[:, 'date'] >= split_from]
        if train_df.shape[0] == 0:
            continue
        elif test_df.shape[0] == 0:
            continue
        clean = pd.concat([clean, player_df])
    return clean

# def clean_df(data, split_from='2018-10-03', streamlit=False):
#     clean = clean_rookies_retirees(data, split_from=split_from)
#     clean, num_unique = clean_duplicates(clean, streamlit=streamlit)
#     return clean, num_unique

@st.cache
def clean_df(data, split_from='2018-10-03', streamlit=False):
    clean = clean_rookies_retirees(data, split_from=split_from)
    clean = clean_duplicates(clean, streamlit=streamlit)
    clean = remove_small_ts(clean, split_from)
    return clean

def sort_dates(data):
    data_df = pd.DataFrame()
    for player_name in data['name'].unique():
        player_df = data[data['name'] == player_name]
        player_df = player_df.sort_values('date')
        data_df = pd.concat([data_df, player_df])
    return data_df

def remove_small_ts(data, split_from='2018-10-03'):
    output_df = pd.DataFrame()
    st.write(data.keys())
    for player_name in data['name'].unique():
        player_df = data[data['name'] == player_name]
        player_test = player_df[player_df['date'] >= split_from]
        player_df = player_df[player_df['date'] < split_from]
        if player_df.shape[0] < 82:
            continue
        player_df = pd.concat([player_df, player_test])
        output_df = pd.concat([output_df, player_df])
    return output_df

def subset_df(data, column_list=None):
    if column_list is None:
        column_list = ['name', 'date', 'cumStatpoints']
    data = data.loc[:, column_list]
    return data

def split_train_test(data, split_from='2018-10-03'):
    train = data.loc[data.loc[:, 'date'] < split_from]
    test = data.loc[data.loc[:, 'date'] >= split_from]
    return train, test

# def prep_df(data, split_from='2018-10-03', column_list=None, streamlit=False, scale=True, target_output_df=False):
#     if column_list is None:
#         column_list = ['name', 'date', 'cumStatpoints']
#     data, num_unique = clean_df(data, split_from=split_from, streamlit=streamlit)
#     if target_output_df:
#         targets, targets_meta, target_df = assemble_target(data, scale=scale, df_output=target_output_df)
#     else:
#         targets, targets_meta = assemble_target(data, scale=scale, df_output=target_output_df)
#     train, test = split_train_test(data)
#     train = train.loc[:, column_list]
#     test = test.loc[:, column_list]
#     if target_output_df:
#         return train, test, num_unique, targets, targets_meta, target_df
#     return train, test, num_unique, targets, targets_meta

def prep_df(data, roster, split_from='2018-10-03', column_list=None, streamlit=False, scale=True, multivar=True):
    st.write('Prepping DF...')
    # st.dataframe(data.head())
    data = data.sort_values('date')
    if column_list is None:
        column_list = ['name', 'date', 'gameNumber', 'cumStatpoints']
    data = clean_df(data, split_from=split_from, streamlit=streamlit)
    data = sort_dates(data)
    # st.dataframe(data.head())
    if multivar:
        # data = remove_small_ts(data, split_from)
        stat_cat_features = assemble_static_cat(data, roster)
        dyn_cat_features = assemble_dynamic_cat(data, roster)
        dyn_real_features, dyn_real_features_meta = days_since_last_game(data)
    targets, targets_meta = assemble_target(data, scale=scale)
    train, test = split_train_test(data)
    # st.dataframe(train.head())
    st.write(column_list)
    train = train.loc[:, column_list]
    # st.dataframe(train.head())
    test = test.loc[:, column_list]
    return train, test, targets, targets_meta, stat_cat_features, dyn_cat_features, dyn_real_features, dyn_real_features_meta

def encode_roster(feature_df, roster):
    position_map = {'C': 1, 'L': 2, 'R': 3, 'D': 4}
    roster = roster.replace({'position': position_map})
    # st.dataframe(roster)
    output_df = pd.DataFrame()
    for player_name in feature_df['name'].unique():
        player_df = feature_df[feature_df['name'] == player_name]
        player_position = roster[roster['name'] == player_name]['position'].iloc[0]
        player_df['position'] = player_position
        output_df = pd.concat([output_df, player_df])
    return output_df

def assemble_dynamic_cat(data, roster, feature_list=['teamId', 'opponentId'], position=False):
    features = data.copy()
    output = []
    features = features.loc[:, ['name', 'date'] + feature_list]
    # features = encode_roster(features, roster)
    for player_name in features['name'].unique():
        player_df = features[features['name'] == player_name]
        # player_df = player_df[['position'] + feature_list]
        player_df = player_df[feature_list]
        output.append(player_df.values.reshape(2, -1))
    return output

def assemble_static_cat(data, roster):
    output = []
    position_map = {'C': 1, 'L': 2, 'R': 3, 'D': 4}
    roster = roster.replace({'position': position_map})
    for player_name in data['name'].unique():
        position = roster[roster['name'] == player_name]['position'].values.item(0)
        output.append([position])
    return output

def days_since_last_game(data, scale=True):
    date_df = data[['date', 'name']]
    date_df['date'] = pd.to_datetime(date_df['date'])
    output = []
    if scale:
        output_meta = pd.DataFrame()
    for player_name in date_df['name'].unique():
        player_df = date_df[date_df['name'] == player_name]
        date_diff = player_df['date']
        date_diff = date_diff.diff()
        date_diff.iloc[0] = pd.Timedelta('187 days 00:00:00')
        date_diff = date_diff.dt.days.values.reshape(-1, 1)
        if scale:
            meta_dict = {'name': player_name}
            scaler = preprocessing.MinMaxScaler()
            date_diff = scaler.fit_transform(date_diff)
            meta_dict['min'] = scaler.min_
            meta_dict['scale'] = scaler.scale_
            meta = pd.DataFrame(meta_dict)
            output_meta = pd.concat([output_meta, meta])
            date_diff = date_diff.tolist()
            # st.write(len(target))
            date_diff = np.array(list(itertools.chain.from_iterable(date_diff))).reshape(1, -1)
            # st.write(target)
        # player_df['dslg'] = date_diff
        output.append(date_diff)
    return output, output_meta

# def generate_metadata(train_data, test_data, num_unique, index=None):
#     # prediction_lengths = [test_data.loc[test_data.loc[:, 'name'] == name].shape[0]
#     #                                     for name in test_data.loc[:, 'name'].unique()]
#     if index is None:
#         if 'date' in train_data.columns:
#             index = 'date'
#         elif 'gameNumber' in train_data.columns:
#             index = 'gameNumber'
#     if index == 'date':
#         data_meta = {'num_series': num_unique,
#                     'num_steps': [train_data.loc[train_data.loc[:, 'name'] == name] \
#                                 .shape[0] for name in train_data.loc[:, 'name'].unique()],
#                     'prediction_length': 82,
#                     'freq': '1D',
#                     'start': [pd.Timestamp(train_data.loc[train_data.loc[:, 'name'] == name] \
#                                             .loc[train_data.loc[train_data.loc[:, 'name'] == name] \
#                                             .index[0], 'date'], freq='1D')
#                             for name in train_data.loc[:, 'name'].unique()]
#                     }
#     elif index == 'gameNumber':
#         data_meta = {'num_series': num_unique,
#                     'num_steps': [train_data.loc[train_data.loc[:, 'name'] == name] \
#                                   .shape[0] for name in train_data.loc[:, 'name'].unique()],
#                     'prediction_length': 82,
#                     'freq': '1D',
#                     'start': [train_data.loc[train_data.loc[:, 'name'] == name] \
#                                         .loc[train_data.loc[train_data.loc[:, 'name'] == name] \
#                                         .index[0], 'gameNumber']
#                               for name in train_data.loc[:, 'name'].unique()]
#                     }
#     return data_meta

def generate_metadata(train_data, test_data, index=None):
    if index is None:
        if 'date' in train_data.columns:
            index = 'date'
        elif 'gameNumber' in train_data.columns:
            index = 'gameNumber'
    prediction_lengths = [test_data.loc[test_data.loc[:, 'name'] == name].shape[0]
                         for name in test_data.loc[:, 'name'].unique()]
    num_unique = len(train_data['name'].unique())
    if index == 'date':
        data_meta = {'num_series': num_unique,
                    'num_steps': [train_data.loc[train_data.loc[:, 'name'] == name] \
                                .shape[0] for name in train_data.loc[:, 'name'].unique()],
                    'prediction_length': prediction_lengths,
                    'freq': '1D',
                    'start': [pd.Timestamp(train_data.loc[train_data.loc[:, 'name'] == name] \
                                            .loc[train_data.loc[train_data.loc[:, 'name'] == name] \
                                            .index[0], 'date'], freq='1D')
                            for name in train_data.loc[:, 'name'].unique()]
                    }
    elif index == 'gameNumber':
        data_meta = {'num_series': num_unique,
                    'num_steps': [train_data.loc[train_data.loc[:, 'name'] == name] \
                                  .shape[0] for name in train_data.loc[:, 'name'].unique()],
                    'prediction_length': prediction_lengths,
                    'freq': '1D',
                    'start': [train_data.loc[train_data.loc[:, 'name'] == name] \
                                        .loc[train_data.loc[train_data.loc[:, 'name'] == name] \
                                        .index[0], 'gameNumber']
                              for name in train_data.loc[:, 'name'].unique()]
                    }
    return data_meta

# def assemble_target(data, feature='cumStatpoints', stand=False, scale=True, df_output=False):
#     targets = []
#     targets_meta = pd.DataFrame()
#     # if stand:
#     #     standardizer = preprocessing.StandardScaler()
#     # if scale:
#     #     scaler = preprocessing.MinMaxScaler()
#     if df_output:
#         targets_df = pd.DataFrame()
#     for player_name in data.loc[:, 'name'].unique():
#         meta_dict = {'name':player_name}
#         player_df = data.loc[data.loc[:, 'name'] == player_name]
#         if not stand and not scale:
#             target = player_df.loc[:, feature].values.tolist()
#         else:
#             target = player_df.loc[:, feature].values.reshape(-1, 1)
#             if stand:
#                 standardizer = preprocessing.StandardScaler()
#                 target = standardizer.fit_transform(target)
#                 meta_dict['mean'] = standardizer.mean_
#                 meta_dict['std'] = standardizer.scale_
#             if scale:
#                 scaler = preprocessing.MinMaxScaler()
#                 target = scaler.fit_transform(target)
#                 meta_dict['min'] = scaler.min_
#                 meta_dict['scale'] = scaler.scale_
#             target = target.tolist()
#             target = list(itertools.chain.from_iterable(target))
#         if df_output:
#             target_df = pd.DataFrame({'name':player_name,
#                                       'gameNumber': player_df.loc[:, 'gameNumber'],
#                                       feature: target})
#             targets_df = pd.concat([targets_df, target_df])
#         targets.append(target)
#         if stand or scale:
#             meta = pd.DataFrame(meta_dict)
#             targets_meta = pd.concat([targets_meta, meta])
#     targets_meta = targets_meta.reset_index(drop=True)
#     if df_output:
#         return targets, targets_meta, targets_df
#     return targets, targets_meta

def assemble_target(data, feature='cumStatpoints', stand=False, scale=True):
    targets = []
    targets_meta = pd.DataFrame()
    # if stand:
    #     standardizer = preprocessing.StandardScaler()
    # if scale:
    #     scaler = preprocessing.MinMaxScaler()
    for player_name in data.loc[:, 'name'].unique():
        meta_dict = {'name':player_name}
        player_df = data.loc[data.loc[:, 'name'] == player_name]
        if not stand and not scale:
            target = player_df.loc[:, feature].values.tolist()
        else:
            target = player_df.loc[:, feature].values.reshape(-1, 1)
            if stand:
                standardizer = preprocessing.StandardScaler()
                target = standardizer.fit_transform(target)
                meta_dict['mean'] = standardizer.mean_
                meta_dict['std'] = standardizer.scale_
            if scale:
                scaler = preprocessing.MinMaxScaler()
                target = scaler.fit_transform(target)
                meta_dict['min'] = scaler.min_
                meta_dict['scale'] = scaler.scale_
            target = target.tolist()
            # st.write(len(target))
            target = list(itertools.chain.from_iterable(target))
            # st.write(target)
        targets.append(target)
        if stand or scale:
            meta = pd.DataFrame(meta_dict)
            targets_meta = pd.concat([targets_meta, meta])
    targets_meta = targets_meta.reset_index(drop=True)
    return targets, targets_meta

def run_model(data_train,
              data_meta,
              save_path,
              num_epochs = 50,
              lr=1e-3,
              batch_size=64):
    estimator = DeepAREstimator(freq=data_meta['freq'],
                                prediction_length=data_meta['prediction_length'],
                                trainer=Trainer(batch_size=batch_size,
                                                epochs=num_epochs,
                                                learning_rate=lr,
                                                ctx='cpu',
                                                hybridize=False))
    predictor = estimator.train(data_train)
    predictor.serialize(Path(save_path))
    return predictor



"""
Module for interacting with the NHL's open but undocumented API.
"""
import streamlit as st
import pandas as pd
from pandas.io.json import json_normalize
import requests as rqsts

## data ingestion
def get_seasons(streamlit=False):
    """ returns all seasons on record """
    seasons_response = rqsts.get('https://statsapi.web.nhl.com/api/v1/seasons')
    try:
        seasons_response.raise_for_status()
    except rqsts.exceptions.HTTPError as e:
        if streamlit:
            st.write(e)
        else:
            print(e)
        raise e
    seasons = seasons_response.content
    seasons_df = pd.read_json(seasons)
    seasons_df = json_normalize(seasons_df.seasons)
    seasons_df.set_index('seasonId', inplace=True)
    return seasons_df

def get_current_season():
    season_response = rqsts.get('https://statsapi.web.nhl.com/api/v1/seasons/current')
    season = season_response.content
    season_df = pd.read_json(season)
    season_df = json_normalize(season_df.seasons)
    season_id = season_df.seasonId
    season_start = season_df.regularSeasonStartDate
    season_end = season_df.regularSeasonEndDate
    return season_id, season_start, season_end

def get_teams(streamlit=False):
    """returns all teams FOR THE CURRENT SEASON"""
    teams_response = rqsts.get('https://statsapi.web.nhl.com/api/v1/teams')
    try:
        teams_response.raise_for_status()
    except rqsts.exceptions.HTTPError as e:
        if streamlit:
            st.write(e)
        else:
            print(e)
        raise e
    teams = teams_response.content
    teams_df = pd.read_json(teams)
    teams_df = json_normalize(teams_df.teams)
    return teams_df

def get_schedule(start_date, end_date):
    # teams = get_teams()
    # st.dataframe(teams)
    # output_df = pd.DataFrame()
    schedule_response = rqsts.get('https://statsapi.web.nhl.com/api/v1/schedule?startDate={0}&endDate={1}'.format(start_date, end_date))
    schedule = schedule_response.content
    schedule = pd.read_json(schedule)
    schedule = json_normalize(schedule.dates)
    output_df = pd.DataFrame()
    for game in schedule.games:
        game_df = json_normalize(game)
        output_df = pd.concat([output_df, game_df])
    return output_df

def get_roster(team_id=22, season_id=20182019, streamlit=False):
    """returns roster for given team and season"""
    roster_response = rqsts.get(('https://statsapi.web.nhl.com/api/'
                                 'v1/teams/{0}'
                                 '?expand=team.roster&season={1}') \
                                 .format(team_id, season_id))
    try:
        roster_response.raise_for_status()
    except rqsts.exceptions.HTTPError as e:
        if streamlit:
            st.write(e)
        else:
            print(e)
        raise e
    roster = roster_response.content
    roster_df = pd.read_json(roster)
    roster_df = json_normalize(roster_df.teams)
    roster_list = roster_df['roster.roster'][0]
    roster_df = pd.DataFrame() # generate df to be filled
    for person in roster_list: # populate the df with desired info for each player
        person_info = person['person']
        person_position = person['position']
        player_dict = {'name': person_info['fullName'],
                       'position': person_position['code']}
        player_df = pd.DataFrame(player_dict, index=[person_info['id']])
        # TODO: should change to not extract as int64
        roster_df = roster_df.append(player_df)
    return roster_df

def merge_team_rosters(team_id=22, season_id_list=None, streamlit=False):
    """returns a roster that includes all players
       that played for a team across the seasons provided"""
    if season_id_list is None:
        season_id_list = [20152016, 20162017, 20172018, 20182019]
    merged_roster_df = pd.DataFrame()
    for season in season_id_list:
        try:
            roster_df = get_roster(team_id=team_id, season_id=season)
        except rqsts.exceptions.HTTPError as e:
            if streamlit:
                st.write(e)
            else:
                print(e)
            continue
        merged_roster_df = merged_roster_df.append(roster_df)
    merged_roster_df.drop_duplicates(inplace=True)
    return merged_roster_df

def get_player_season_game_stats(player_id=8477934, season_id=20182019, streamlit=False):
    """gets the game-by-game stats for a given player and season"""
    stats_response = rqsts.get('https://statsapi.web.nhl.com/api/'
                               'v1/people/{0}/stats?stats=gameLog&season={1}' \
                                .format(player_id, season_id))
    try:
        stats_response.raise_for_status()
    except rqsts.exceptions.HTTPError as e:
        if streamlit:
            st.write(e)
        else:
            print(e)
        raise e
    stats = stats_response.content
    stats_df = pd.read_json(stats)
    stats_df = json_normalize(stats_df.stats)
    stats_df = stats_df.splits
    stats_array = stats_df.array
    stats_list = stats_array[0]
    stats_df = pd.DataFrame() # generate stat df to be filled
    for game in stats_list:
        # import and transpose df from returned game dict (json)
        game_df = pd.DataFrame.from_dict(game).transpose()
        clean_df = pd.DataFrame() # generate game df to be filled
        for stat_type, stat_series in game_df.iterrows():
            if stat_type != 'stat': # the 'stat' stat type contains non-unique but desired values
                try:
                    stat_series.drop_duplicates(inplace=True)
                except SystemError:
                    pass
            stat_series.dropna(inplace=True) # clean out NaN
            # transpose the series so it fits properly into our df
            stat_df = pd.DataFrame(stat_series).transpose()
            # rename columns to prevent collision
            new_columns = []
            if len(stat_df.columns) != 1:
                new_columns = [stat_type + column.capitalize()
                               for column in stat_df.columns]
            if not new_columns:
                new_columns = stat_df.index # use the index if there are no new columns
            stat_df.reset_index(drop=True, inplace=True)
            stat_df.columns = new_columns # rename columns
            clean_df = pd.concat([clean_df, stat_df], axis=1) # add the game to the df
        game_df = clean_df.drop('gameContent', axis=1) # replace the dirty df with our clean one
        # set the indices to be the unique game identifier
        game_df.set_index('gameGamepk', inplace=True)
        stats_df = stats_df.append(game_df)
    return stats_df

def get_combined_player_season_game_stats(player_id=8477934, season_id_list=None):
    """returns player game-by-game stats across multiple seasons"""
    # TODO: add function to set whether each season should be individual cumulative totals or cumulative across all seasons
    if season_id_list is None:
        season_id_list = [20152016, 20162017, 20172018, 20182019]
    full_df = pd.DataFrame()
    for season_id in season_id_list:
        season_df = get_player_season_game_stats(player_id=player_id, season_id=season_id)
        full_df = pd.concat([full_df, season_df])
    # drop all duplicate entries (resulting from playing for multiple teams)
    full_df.drop_duplicates(subset='date', keep='first', inplace=True)
    return full_df

def augment_player_dataframe(player_df, cumulative_stat_list=None):
    """generates cumulative totals of stats"""
    if cumulative_stat_list is None:
        cumulative_stat_list = ['statPoints']
    augmented_df = player_df
    augmented_df.sort_index(inplace=True) # make sure everything is in order
    for stat in cumulative_stat_list:
        try:
            stat_series = augmented_df.loc[:, stat] # grab a stat
        # TODO: verify why there are no points for these players; THINK its because I asked for seasons that didn't exist. still necessary?
        except KeyError:
            stat_series = pd.DataFrame({'cum'+stat.capitalize():
                                        [None for _ in range(len(augmented_df))]})
        stat_series = stat_series.cumsum()
        try:
            stat_series.name = 'cum' + stat_series.name.capitalize() # rename the stat column
        except AttributeError:
            pass
        augmented_df = pd.concat([augmented_df, stat_series], axis=1)
    augmented_df.insert(0, 'gameNumber',
                        [i+1 for i in range(len(augmented_df))]) # add game numbers column
    return augmented_df

def get_player_name_position(player_id=8477934, streamlit=False):
    """returns CURRENT basic player name and position"""
    player_response = rqsts.get('https://statsapi.web.nhl.com/api/v1/people/{0}/'.format(player_id))
    try:
        player_response.raise_for_status()
    except rqsts.exceptions.HTTPError as e:
        if streamlit:
            st.write(e)
        else:
            print(e)
        raise e
    player = player_response.content
    player = pd.read_json(player)
    player = player.people[0]
    player_name = player['fullName']
    player_position = player['primaryPosition']['code']
    return player_name, player_position

def assemble_multiplayer_stat_dataframe(player_id_list=None, season_id_list=None,
                                        stat_list=None, shape='cols'):
    """returns game-by-game stats for given list of players across given seasons
       (can specify sspecific stats for smaller returns)"""
    if stat_list is None:
        stat_list = ['cumStatpoints']
    if player_id_list is None:
        player_id_list = [8477934, 8476356, 8473468]
    if season_id_list is None:
        season_id_list = [20152016, 20162017, 20172018, 20182019]
    multiplayer_df = pd.DataFrame() # TODO: add a progress bar
    for player_id in player_id_list:
        player_name, player_position = get_player_name_position(player_id)
        if player_position == 'G':
            continue # don't include goalies
        if len(season_id_list) == 1: # handle single or multiple season input lists
            player_df = augment_player_dataframe(
                get_player_season_game_stats(
                    player_id=player_id, season_id=season_id_list[0]))
        else:
            player_df = augment_player_dataframe(
                get_combined_player_season_game_stats(
                    player_id=player_id, season_id_list=season_id_list))
        if stat_list: # if a specific stat is given, only grab that
            player_small_df = player_df \
                                       .loc[:, ['date', 'gameNumber'] \
                                                + stat_list] # keep the useful indices
        else:
            player_small_df = player_df # grab it all otherwise
        player_small_df.reset_index(drop=True, inplace=True) # get rid of messy index
        try:
            player_small_df.insert(0, 'name', [player_name for _ in range(len(player_small_df))])
        except ValueError: # still don't know why this is thrown sometimes
            st.dataframe(player_small_df)
            player_small_df.insert(0, 'errorName',
                                   [player_name for _ in range(len(player_small_df))])
        # player_small_df.set_index('gameNumber', inplace=True)
        # player_small_df.rename(columns={stat: player_name}, inplace=True)
        multiplayer_df = pd.concat([multiplayer_df, player_small_df], axis=0)
        # save temp results
        # multiplayer_df.to_csv('../../data/temp/'
        #                       'assemble_multiplayer_stat_dataframe_TEMP.csv')
    multiplayer_df.reset_index(drop=True, inplace=True)
    if shape == 'rows': # TODO: fix this when required to feed data in a row-per-player fashion
        multiplayer_df = multiplayer_df.transpose()
        # multiplayer_df.set_index(player_id_list, inplace=True)
        try:
            multiplayer_df.insert(0, 'name', multiplayer_df.index)
        except ValueError:
            multiplayer_df.insert(0, 'errorName', multiplayer_df.index)
        multiplayer_df.insert(0, 'playerId', player_id_list)
        multiplayer_df.set_index('playerId', inplace=True)
    return multiplayer_df

def get_all_rosters(season_id_list=None, streamlit=False):
    """returns roster of all players across a given list of seasons"""
    if season_id_list is None:
        season_id_list = [20152016, 20162017, 20172018, 20182019]
    teams_df = get_teams()
    team_ids = teams_df.loc[:, 'id']
    full_roster_df = pd.DataFrame()
    for team_id in team_ids:
        # TODO: save player id's
        team_full_roster = merge_team_rosters(team_id=team_id,
                                              season_id_list=season_id_list,
                                              streamlit=streamlit)
        full_roster_df = pd.concat([full_roster_df, team_full_roster])
    return full_roster_df

## HERE BE DRAGONS
# The following functions are either not complete or broken.
# def generate_games_df(schedule_df):
#     """generate dataframe of ALL games in a schedule"""
#     games_df = pd.DataFrame()
#     games_series = schedule_df.games
#     st.dataframe(games_series)
#     for day in games_series:
#         for game in day:
#             if game['gameType'] == 'R':
#                 st.dataframe(game)
#                 game_id = game['gamePk']
#                 game_teams = game['teams']
#                 away_team = game_teams['away']
#                 home_team = game_teams['home']
#                 st.write(home_team.keys())
#                 return game_id, home_team, away_team
#                 # games_df.append(game, ignore_index=True)
#     st.dataframe(games_df)
#     return None, None, None

# def get_season_schedule(seasons_df=get_seasons(), season_id=20182019, streamlit=False):
#     """get schedule for a season"""
#     season_series = seasons_df.loc['{}'.format(season_id), :]
#     st.table(season_series)
#     season_start = season_series['regularSeasonStartDate']
#     season_end = season_series['regularSeasonEndDate']
#     st.text('Season ID: {0}\nSeason start: {1}\nSeason end: {2}' \
#             .format(season_id, season_start, season_end))
#     schedule_response = rqsts.get('https://statsapi.web.nhl.com/api/'
#                                   'v1/schedule?startDate={0}&endDate={1}' \
#                                   .format(season_start, season_end))
#     try:
#         schedule_response.raise_for_status()
#     except rqsts.exceptions.HTTPError as e:
#         if streamlit:
#             st.write(e)
#         else:
#             print(e)
#     schedule = schedule_response.content
#     schedule_df = pd.read_json(schedule)
#     schedule_df = json_normalize(schedule_df.dates)
#     schedule_df.set_index('date', inplace=True)
#     # schedule_df = json_normalize(schedule_df.games)
#     # st.dataframe(schedule_df.games)
#     generate_games_df(schedule_df)
#     return schedule_df



"""
Module to visualize model inputs, outputs, and performance
"""
from datetime import datetime as dt
import streamlit as st
import numpy as np
import pandas as pd
from bokeh.layouts import column
from bokeh.models import ColumnDataSource, RangeTool, Span, HoverTool, FixedTicker, PrintfTickFormatter
from  bokeh.models.ranges import Range1d
from bokeh.plotting import figure
from scipy.stats.kde import gaussian_kde
from scipy.stats import ttest_ind
import colorcet as cc

def calculate_predictions(data, results, player_name='Leon Draisaitl', target='cumStatpoints'):
    """ calculate predictions from residuals and the original data"""
    test_results = results.loc[player_name, :]
    test_residuals = test_results.testResiduals
    train_residuals = test_results.trainResiduals
    test_real = data[data['name'] == player_name].loc[:, ['date', target]]
    full_residuals = pd.concat([train_residuals, test_residuals], axis=0)
    full_residuals.reset_index(inplace=True, drop=True)
    full_residuals.columns = ['residuals']
    test_real.reset_index(inplace=True, drop=True)
    full_frame = pd.concat([test_real, full_residuals], axis=1)
    full_frame['date'] = pd.to_datetime(full_frame['date'])
    full_frame.drop_duplicates(subset='date', keep='first', inplace=True)
    full_frame.set_index('date', drop=False, inplace=True)
    try:
        full_frame['predictions'] = full_frame \
                                        .apply(lambda row: row.cumStatpoints - row.residuals, axis=1)
    except Exception as e:
        raise e
    full_frame.loc[:, 'name'] = player_name
    return full_frame

def calculate_errors(residuals, start_date='2018-10-03', end_date=None):
    """ Calculates errors based on residuals """
    if start_date:
        residuals = residuals.loc[residuals.loc[:, 'date'] >= start_date]
    if end_date:
        residuals = residuals.loc[residuals.loc[:, 'date'] < end_date]
    num_residuals = len(residuals)
    residuals = pd.DataFrame(residuals.loc[:, 'residuals'])
    mfe = (residuals.sum() / num_residuals).tolist()[0]
    mae = (residuals.abs().sum() / num_residuals).tolist()[0]
    rmse = ((residuals.pow(2).sum() / num_residuals).pow(0.5)).tolist()[0]
    residuals = residuals.values
    residuals = [value.item() for value in residuals]
    return mfe, mae, rmse

def calculate_residuals(predictions, start_date='2018-10-03', end_date=None):
    if start_date:
        predictions = predictions.loc[predictions.loc[:, 'date'] >= start_date]
    if end_date:
        predictions = predictions.loc[predictions.loc[:, 'date'] < end_date]
    residuals = predictions.loc[:, 'cumStatpoints'] - predictions.loc[:, 'predictions']
    residuals = pd.DataFrame({'date':predictions.loc[:, 'date'],
                              'name':predictions.loc[:, 'name'],
                              'residuals':residuals})
    return residuals

def calculate_error_df(predictions, drift_loc, start_date='2018-10-03', end_date=None, use_mase=True):
    if use_mase:
        drift_df = pd.read_csv(drift_loc)
    error_df = pd.DataFrame()
    for player in predictions.loc[:, 'name'].unique():
        mfe, mae, rmse = calculate_errors(
                            calculate_residuals(predictions.loc[predictions.loc[:, 'name'] == player], 
                                                start_date=start_date,
                                                end_date=end_date),
                            start_date=start_date,
                            end_date = end_date)

        if use_mase:

            player_drift = drift_df.loc[drift_df.loc[:, 'name'] == player]
            drift_residuals = calculate_residuals(player_drift, 
                                                    start_date='2018-10-03',
                                                    end_date=None)
            _, drift_mae, _ = calculate_errors(
                                drift_residuals,
                                start_date='2018-10-03',
                                end_date = None)
            if drift_mae != 0:
                mase = mae / drift_mae
            else:
                mase = float('NaN')
        if use_mase:
            player_df = pd.DataFrame({'name':player,
                                    'mase':mase,
                                    'mfe':mfe,
                                    'mae':mae,
                                    'rmse':rmse}, index=[player])
        else:
            player_df = pd.DataFrame({'name':player,
                                    'mfe':mfe,
                                    'mae':mae,
                                    'rmse':rmse}, index=[player])
        error_df = pd.concat([error_df, player_df])
        error_df = error_df.reset_index(drop=True)
    return error_df

def generate_error_df(predictions_file_list, data_loc, drift_loc, metric='mase'):
    output_df = pd.DataFrame()
    for prediction_file in predictions_file_list:
        if metric == 'mase' and 'drift' in prediction_file:
            continue
        if prediction_file.split('.')[-1] == 'p':
            data = pd.read_csv(data_loc)
            results = pd.read_pickle(prediction_file)
            results.loc[:, 'name'] = results.index
            predictions = pd.DataFrame()
            for player in  results.loc[:, 'name'].unique():
                try:
                    player_prediction = calculate_predictions(data, results, player_name=player)
                except AttributeError:
                    continue
                predictions = pd.concat([predictions, player_prediction])
        else:
            predictions = pd.read_csv(prediction_file)
        if 'drift' in prediction_file:
            error_df = calculate_error_df(predictions, drift_loc, start_date='2018-10-03', end_date=None)
        else:
            error_df = calculate_error_df(predictions, drift_loc)
        error_df = pd.DataFrame({'name.{}'.format(prediction_file.split('/')[-1].split('.')[0]): error_df.loc[:, 'name'],
                                 prediction_file.split('/')[-1].split('.')[0]: error_df.loc[:, metric]})
        output_df = pd.concat([output_df, error_df], axis=1)
    return output_df

def test_errors(error_df, dist1='arima_results_m3', dist2='deepar_truncated_results_default', report=False):
    t, p = ttest_ind(error_df.loc[:, dist1], error_df.loc[:, dist2], equal_var=False, nan_policy='omit')
    if report:
        st.text('{0} model mean MASE: {1:.3f} (std.: {2})\n{3} model mean MASE: {4:.3f} (std.: {5})' \
                 .format(dist1,
                         error_df.loc[:, dist1].mean(),
                         error_df.loc[:, dist1].std(),
                         dist2,
                         error_df.loc[:, dist2].mean(),
                         error_df.loc[:, dist2].std()))
    return t, p

def calculate_par(player, joe):
    foo = None
    # st.dataframe(player)
    # st.dataframe(joe)
    return

def assemble_diagnoses(data, errors):
    return

def return_intervals(results, player_name='Leon Draisaitl'):
    """ return the prediction intervals from a results dataframe"""
    lows = results.loc[player_name, 'intervalLow']
    highs = results.loc[player_name, 'intervalHigh']
    try:
        intervals = pd.DataFrame({'low':lows, 'high':highs})
    except ValueError:
        intervals = pd.DataFrame({'low':lows.tolist(), 'high':highs.tolist()})
    return intervals

def plot_actual_predictions_series(data,
                                   results,
                                   model_errors,
                                   joe=None,
                                   par=False,
                                   model='arima',
                                   target='cumStatpoints',
                                   metric='mase',
                                   player_name='Leon Draisaitl',
                                   deepar_model_name='deepar_truncated_results_unit_s_ne300_lr1e-3_bs64_nl3_cl3.csv'):
    """ plots the real and predicted time series along with confidence intervals for a player"""
    if results is None:
        model = 'deepar'
    if model == 'arima':
        player_df = data.loc[data.loc[:, 'name'] == player_name]
        player_errors = model_errors.loc[model_errors.loc[:, 'name.arima_results_m3'] == player_name]
        try:
            player_error = player_errors['arima_results_m3'].values.astype(float).item()
        except ValueError:
            st.write('Unfortunately, Auto-ARIMA was unable to find a suitable model for this player.')
            return
        series_dataframe = calculate_predictions(data, results, player_name=player_name, target=target)
        intervals = return_intervals(results, player_name)
        dates = series_dataframe.index.values.astype(np.datetime64)
        joe = joe.loc[joe.loc[:, 'date'].isin(player_df.date)]
        joe = joe.set_index('date')
        compare_dataframe = joe.loc[:, 'cumStatpoints']
        compare_dataframe = compare_dataframe.sort_index()
    elif model == 'deepar':
        player_df = data.loc[data.loc[:, 'name'] == player_name]
        if model_errors is not None:
            player_errors = model_errors.loc[model_errors.loc[:, 'name.{}'.format(deepar_model_name)] == player_name]
            try:
                player_error = player_errors[deepar_model_name].values.astype(float).item()
            except ValueError:
                st.write('This player was not able to be evaluated.')
        else:
            player_df = player_df.sort_values('gameNumber', ascending=True)
            player_df = player_df.set_index('gameNumber')
            # player_df.dropna(axis=1, subset=['predictions'])
        # _, _, rmse = calculate_errors(calculate_residuals(player_df))
        if model_errors is not None:
            series_dataframe = player_df.loc[:, [target, 'predictions']]
        else:
            series_dataframe = player_df.loc[:, 'predictions']
        series_dataframe = series_dataframe.sort_index()
        intervals = player_df.loc[:, ['high', 'low']].dropna()
        intervals = intervals.sort_index()
        if model_errors is not None:
            dates = series_dataframe.index.values.astype(np.datetime64)
        else:
            gameNumbers = series_dataframe.index.values.astype(int)
        # st.write(dates)
        if model_errors is not None:
            joe = joe.loc[joe.loc[:, 'date'].isin(series_dataframe.index)]
            joe = joe.set_index('date')
            compare_dataframe = joe.loc[:, 'cumStatpoints']
            compare_dataframe = compare_dataframe.sort_index()
        if par:
            calculate_par(series_dataframe, compare_dataframe)
    elif data is None and results is None:
        player_name = 'Joe Schmo'
        joe = joe.set_index('date')
        series_dataframe = joe.loc[:, 'cumStatpoints']
        series_dataframe = series_dataframe.sort_index()
        dates = series_dataframe.index.values.astype(np.datetime64)
    start_date = dt.strptime('2018-10-03', '%Y-%m-%d')

    if model_errors is not None:
        # st.dataframe(model_errors)
        real_source = ColumnDataSource(data=dict(date=dates, points=series_dataframe[target]))
        compare_source = ColumnDataSource(data=dict(date=dates, points=compare_dataframe))
        pred_source = ColumnDataSource(data=dict(date=dates, points=series_dataframe['predictions']))
        interval_dates = dates[-intervals.shape[0]:].reshape(-1, 1)
        interval_dates = np.hstack((interval_dates, interval_dates))
    else:
        pred_source = ColumnDataSource(data=dict(gameNumber=gameNumbers, points=series_dataframe))
        interval_games = gameNumbers[-intervals.shape[0]:].reshape(-1, 1)
        interval_games = np.hstack((interval_games, interval_games))

    if model_errors is not None:
        y_max = max([series_dataframe.loc[:, target].max(), joe.loc[:, target].max()]) + 50
    else:
        try:
            y_min = series_dataframe.values[0]
        except IndexError:
            st.write('This player was not able to be predicted.')
            return
        y_max = series_dataframe.max() + 50

    if model == 'arima':
        player_line = figure(title=('{0}({1},{2},{3})({4},{5},{6},{7}) ({8}: {9:.3f})') \
                                    .format(player_name,
                                            results.loc[player_name, 'p'],
                                            results.loc[player_name, 'd'],
                                            results.loc[player_name, 'q'],
                                            results.loc[player_name, 'P'],
                                            results.loc[player_name, 'D'],
                                            results.loc[player_name, 'Q'],
                                            3, # TODO: undo hardcoding
                                            metric.upper(),
                                            player_error), # TODO: change to MASE
                            plot_height=300,
                            plot_width=800,
                            tools="xpan",
                            toolbar_location='above',
                            x_axis_type="datetime",
                            x_axis_location="below",
                            x_range=(start_date, dates[-1]),
                            y_range=(0, y_max),
                            x_axis_label='Game Date',
                            y_axis_label='Cumulative Points',
                            background_fill_color="#efefef"
                            )
    else:
        if model_errors is not None:
            player_line = figure(title=('{0} ({1}: {2:.3f})'.format(player_name, metric.upper(), player_error)), # TODO: add error
                                plot_height=300,
                                plot_width=800,
                                tools="xpan",
                                toolbar_location='above',
                                x_axis_type="datetime",
                                x_axis_location="below",
                                x_range=(start_date, dates[-1]),
                                y_range=(0, y_max),
                                x_axis_label='Game Date',
                                y_axis_label='Cumulative Points',
                                background_fill_color="#efefef"
                                )
        else:
            player_line = figure(title=('{0}'.format(player_name)), # TODO: add error
                                plot_height=300,
                                plot_width=800,
                                tools="xpan",
                                toolbar_location='above',
                                x_axis_location="below",
                                x_range=(0, 82),
                                y_range=(y_min, y_max),
                                x_axis_label='Game Date',
                                y_axis_label='Cumulative Points',
                                background_fill_color="#efefef"
                                )
    # elif data is None and results is None:
    #     player_line = figure(title=('{0}'.format(player_name)), # TODO: add error
    #                          plot_height=300,
    #                          plot_width=800,
    #                          tools="xpan",
    #                          toolbar_location='above',
    #                          x_axis_type="datetime",
    #                          x_axis_location="below",
    #                          x_range=(dates[0], dates[-1]),
    #                          background_fill_color="#efefef"
    #                          )

    # hover_tool = HoverTool(tooltips=[("date", "@date"),
    #                                  ("points", "@points")
    #                                 ],
    #                        mode='vline'
    #                        )

    if model_errors is not None:
        player_line.line('date', 'points', source=real_source, line_color='blue', legend='actual')
        player_line.line('date', 'points', source=compare_source, line_color='green', legend='average')
        player_line.circle('date', 'points', source=pred_source, line_color='red', fill_color='red', legend='predicted')
        player_line.varea(x=interval_dates[:, 0],
                y1=intervals.loc[:, 'high'],
                y2=intervals.loc[:, 'low'],
                fill_alpha=0.4,
                color='red',
                legend='predicted')
    else:
        player_line.circle('gameNumber', 'points', source=pred_source, line_color='red', fill_color='red', legend='predicted')
        player_line.varea(x=interval_games[:, 0],
                y1=intervals.loc[:, 'high'],
                y2=intervals.loc[:, 'low'],
                fill_alpha=0.4,
                color='red',
                legend='predicted')



    player_line.legend.location = 'bottom_right'
    player_line.legend.click_policy = 'hide'
    # player_line.add_tools(hover_tool)
    # player_line.toolbar.active_multi = hover_tool
    # player_line.yaxis.axis_label('Cumulative Points')

    test_start = Span(location=start_date,
                      dimension='height', line_color='green',
                      line_dash='dashed', line_width=3)
    player_line.add_layout(test_start)

    if model_errors is not None:
        select = figure(title=("Drag the middle and edges of the"
                            "selection box to change the range above"),
                        plot_height=130,
                        plot_width=800,
                        y_range=player_line.y_range,
                        x_axis_type="datetime",
                        y_axis_type=None,
                        x_axis_label='Game Date',
                        y_axis_label='Cumulative Points',
                        tools="",
                        toolbar_location=None,
                        background_fill_color="#efefef",
                        x_range=(dates[0], dates[-1]))
    else:
        select = figure(title=("Drag the middle and edges of the"
                        "selection box to change the range above"),
                        plot_height=130,
                        plot_width=800,
                        y_range=player_line.y_range,
                        x_axis_type="datetime",
                        y_axis_type=None,
                        x_axis_label='Game Date',
                        y_axis_label='Cumulative Points',
                        tools="",
                        toolbar_location=None,
                        background_fill_color="#efefef",
                        x_range=(0, 82))

    # range_tool = RangeTool(x_range=Range1d(start_date, dates[-1]))
    range_tool = RangeTool(x_range=player_line.x_range)
    range_tool.overlay.fill_color = "navy"
    range_tool.overlay.fill_alpha = 0.2

    if model_errors is not None:
        select.line('date', 'points', source=real_source, line_color='blue')
        select.line('date', 'points', source=compare_source, line_color='green')
        select.circle('date', 'points', source=pred_source, line_color='red', fill_color='red')
    else:
        select.circle('gameNumber', 'points', source=pred_source, line_color='red', fill_color='red')

    select.ygrid.grid_line_color = None
    select.add_tools(range_tool)
    select.toolbar.active_multi = range_tool
    select.add_layout(test_start)

    chart = column(player_line, select)

    st.bokeh_chart(chart)
    # st.write("{}'s dataframe:".format(player_name))
    # st.dataframe(series_dataframe)
    return chart

def ridge(category, data, scale=3):
    return list(zip([category]*len(data), scale*data))

def ridge_plots(model_df):
    models = list(model_df.keys())
    models.reverse()
    for i, model in enumerate(models):
        if 'name' in model:
            models.pop(i)

    palette = [cc.rainbow[i*15] for i in range(len(models))]

    x = np.linspace(-20, 110, 500)

    source = ColumnDataSource(data=dict(x=x))

    p = figure(y_range=models, plot_width=900, x_range=(0, 5), toolbar_location=None)

    for i, model in enumerate(models):
        if 'name' in model:
            continue
        pdf = gaussian_kde(model_df[model].dropna())
        y = ridge(model, pdf(x))
        source.add(y, model)
        p.patch('x', model, color=palette[i], alpha=0.6, line_color="black", source=source)

    p.outline_line_color = None
    p.background_fill_color = "#efefef"

    p.xaxis.ticker = FixedTicker(ticks=list(range(int(model_df.min().min()), 5, 1)))
    # p.xaxis.formatter = PrintfTickFormatter(format="%d%%")

    p.ygrid.grid_line_color = None
    p.xgrid.grid_line_color = "#dddddd"
    p.xgrid.ticker = p.xaxis[0].ticker

    p.axis.minor_tick_line_color = None
    p.axis.major_tick_line_color = None
    p.axis.axis_line_color = None

    p.y_range.range_padding = 2

    base = Span(location=1,
                dimension='height', line_color='black',
                line_dash='dashed', line_width=3)
    p.add_layout(base)

    st.bokeh_chart(p)
    return

## HERE BE DRAGONS
# The following functions are either not complete or broken.
# def get_hists(results_list, metric_list=None, result_names=None, edges_method='min'):
#     """ get the histograms and edges from a set of results """
#     if metric_list is None:
#         metric_list = ['testRmse']
#     if result_names is None:
#         result_names = ['Raw', 'Yeo-Johnson']
#     hists = []
#     edges_list = []
#     for result in results_list:
#         for metric in metric_list:
#             hist, edges = np.histogram(result[metric])
#             hists.append(hist)
#             edges_list.append(edges)
#     edge_min = edges_list[0][-1]
#     edge_max = edges_list[0][-1] # I cheated here
#     edge_loc = None
#     for i, edges in enumerate(edges_list):
#         edge_range = edges[-1] - edges[0]
#         if edges_method == 'max':
#             if edge_range > edge_max:
#                 edge_loc = i
#         else:
#             if edge_range < edge_min:
#                 edge_loc = i
#     if not edge_loc:
#         edge_loc = 0
#     edges = edges_list[edge_loc]
#     st.dataframe(edges)
#     return hists, edges

# def plot_hists(result_list, metric_list=None, result_names=None):
#     """plot a set of histograms from a list of results for the given metrics"""
#     if metric_list is None:
#         metric_list = ['testRmse']
#     if result_names is None:
#         result_names = ['Raw', 'Yeo-Johnson']
#     fig = figure(plot_height=600,
#                  plot_width=600,
#                  title="Histogram of ARIMA RMSE's",
#                  x_axis_label='RMSE',
#                  y_axis_label='# of players')

#     results_df = pd.DataFrame()
#     for result, name in zip(result_list, result_names):
#         metrics_df = pd.DataFrame()
#         for metric in metric_list:
#             metric_df = result[metric]
#             metrics_df = pd.concat([metrics_df, metric_df], axis=1)
#         metrics_df.columns = [column + name for column in metrics_df.columns]
#         results_df = pd.concat([results_df, metrics_df], axis=1)
#         # TODO: I don't think multiple metrics will play nice here...
#         # result_df = pd.concat([result[metric] for metric in metric_list], axis=1)
#     max_range = results_df.max() - results_df.min()
#     st.dataframe(max_range)

#     # TODO: add scaling of axes back in
#     hists, edges = get_hists(metric_list, result_list, result_names, edges_method='max')
#     # TODO: handle colors/length mismatch
#     for result_name, hist, color in zip(result_names, hists, ['blue', 'red', 'green']):
#         fig.quad(bottom=0,
#                  top=hist,
#                  left=edges[:-1],
#                  right=edges[1:],
#                  fill_color=color,
#                  line_color='black',
#                  legend=result_name)

#     fig.legend.location = 'top_right'
#     fig.legend.click_policy = 'hide'

#     st.bokeh_chart(fig)
#     return fig



"""
Module for running ARIMA on players
"""
import time
import streamlit as st
import numpy as np
import pandas as pd
import pmdarima as pm
from sklearn.preprocessing import PowerTransformer

def calculate_errors(residuals):
    """ Calculates errors based on residuals """
    num_residuals = len(residuals)
    mfe = (residuals.sum() / num_residuals).tolist()[0]
    mae = (residuals.abs().sum() / num_residuals).tolist()[0]
    rmse = (residuals.pow(2).sum().pow(0.5)).tolist()[0]
    residuals = residuals.values
    residuals = [value.item() for value in residuals]
    return mfe, mae, rmse

def calculate_test_residuals(prediction_array, test_data):
    """ Calculates test residuals based on prediction and test data """
    prediction_array = prediction_array.reshape(len(test_data), 1)
    test_data = test_data.values
    residuals = np.subtract(test_data, prediction_array)
    residuals = residuals.tolist()
    residuals = pd.DataFrame(residuals)
    return residuals

def player_arima(data,
                 player_name,
                 index='date',
                 feature='cumStatpoints',
                 forecast_from='2018-10-03',
                 transform='none',
                 player_id=None,
                 roster=None,
                 summary=False):
    """ performs Auto-ARIMA on a single player """
    # TODO: add logic for if the player ID is given but not a roster (use function in package)
    if player_id and roster:
        player_name = roster[roster['Unnamed: 0'] == player_id]
    player_df = data[data['name'] == player_name]
    player_df.drop_duplicates(subset='date', keep='first', inplace=True)
    player_train_df = player_df[player_df['date'] < forecast_from]
    player_test_df = player_df[player_df['date'] >= forecast_from]
    player_train_df = player_train_df.loc[:, [index, feature]]
    player_train_df = player_train_df.set_index(index, drop=True)
    if player_train_df.shape[0] == 0:
        st.write('{} is a rookie!'.format(player_name))
        return None
    if transform == 'log':
        # TODO: make this stat agnostic
        player_train_df.loc[:, 'logValues'] = np.log(player_train_df['cumStatpoints'])
    elif transform == 'yj':
        transformer = PowerTransformer()
        transformer.fit(player_train_df.values.reshape(-1, 1))
        player_train_df.loc[:, 'transformedValues'] = transformer \
                                                      .transform(
                                                          player_train_df['cumStatpoints'] \
                                                          .values.reshape(-1, 1))
        player_train_df.drop('cumStatpoints', axis=1, inplace=True)
    player_test_df = player_test_df.loc[:, [index, feature]]
    player_test_df = player_test_df.set_index(index, drop=True)
    # player_train_df = player_train_df[:'2018-10-03']
    # player_test_df = player_test_df['2018-10-03':]
    if player_test_df.shape[0] == 0:
        st.write('{} retired!'.format(player_name))
        return None
    start_time = time.time()
    st.write('Searching ARIMA parameters for {}...'.format(player_name))
    try:
        model = pm.auto_arima(player_train_df,
                              start_p=1,
                              start_q=1,
                              max_p=5,
                              max_q=5,
                              max_d=3,
                              m=3,
                              start_P=0,
                              start_Q=0,
                              seasonal=True,
                              information_criterion='aicc',
                              error_action='ignore',
                              suppress_warnings=True,
                              stepwise=True)
        st.write('Model built, fitting...')
        model.fit(player_train_df)
    except ValueError:
        st.write("{} doesn't have enough data!".format(player_name))
        return None
    except IndexError:
        st.write('Index error for {}'.format(player_name))
        return None
    except:
        st.write('Unhandled error for {}'.format(player_name))
        return None
    predictions, intervals = model.predict(n_periods=player_test_df.shape[0], return_conf_int=True)
    if transform == 'log':
        predictions = np.exp(predictions)
        intervals = np.exp(intervals)
    elif transform == 'yj':
        predictions = transformer.inverse_transform(predictions.reshape(-1, 1))
        low_intervals = transformer.inverse_transform(intervals[:, 0].reshape(-1, 1))
        high_intervals = transformer.inverse_transform(intervals[:, 1].reshape(-1, 1))
    end_time = time.time()
    if transform != 'yj':
        low_intervals = []
        high_intervals = []
        for low, high in intervals:
            low_intervals.append(low)
            high_intervals.append(high)
    prediction_residuals = calculate_test_residuals(predictions, player_test_df)
    if summary:
        st.text(model.summary())
    train_residuals = pd.DataFrame(model.resid())
    train_mfe, train_mae, train_rmse = calculate_errors(train_residuals)
    test_mfe, test_mae, test_rmse = calculate_errors(prediction_residuals)
    model_params = model.get_params()
    p, d, q = model_params['order']
    try:
        P, D, Q, m = model_params['seasonal_order']
    except TypeError:
        st.write('Search failed to find valid options.')
        return None
    st.write("{0}'s Auto-ARIMA({1},{2},{3})({4},{5},{6},{7}) took {8:.3f} seconds." \
             .format(player_name, p, d, q, P, D, Q, m, end_time-start_time))
    results_df = pd.DataFrame({'forecastStart':forecast_from,
                               'aic':model.aic(),
                               'p':p,
                               'd':d,
                               'q':q,
                               'P':P,
                               'D':D,
                               'Q':Q,
                               'm':m,
                               'trainMfe':train_mfe,
                               'trainMae':train_mae,
                               'trainRmse':train_rmse,
                               'trainResiduals':[train_residuals],
                               'testMfe':test_mfe,
                               'testMae':test_mae,
                               'testRmse':test_rmse,
                               'testResiduals':[prediction_residuals],
                               'intervalLow':[low_intervals],
                               'intervalHigh':[high_intervals]},
                              index=[player_name])
    return results_df

def all_player_arima(data, roster, save_loc, transform='none', print_status=False):
    """ performs Auto_ARIMA on all players in a given roster """
    if print_status:
        print('Running Auto-ARIMAs...')
    results = pd.DataFrame()
    for index, player in roster.iterrows():
        if print_status:
            print('Player {}'.format(index))
        player_name = player['name']
        player_results = player_arima(data, player_name=player_name, transform=transform)
        if isinstance(player_results, type(None)):
            st.write('Skipping {}'.format(player_name))
            continue
        st.dataframe(player_results)
        results = pd.concat([results, player_results])
        results.to_pickle(save_loc)
    if print_status:
        print('Done!')
    return results
